/*
 * This file belongs to FreeMiNT. It's not in the original MiNT 1.12
 * distribution. See the file CHANGES for a detailed log of changes.
 * 
 * 
 * Copyright 2005 Patrice Mandin <pmandin@caramail.com>
 * All rights reserved.
 * 
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * 
 * Author: Patrice Mandin <pmandin@caramail.com>
 * Started: 2005-04-29
 * 
 * Please send suggestions, patches or bug reports to me or
 * the MiNT mailing list.
 * 
 * 
 * changes since last version:
 * 
 * 
 * 2005-04-29:	(v0.1)
 * 
 * - initial revision
 * 
 * 
 * bugs/todo:
 * 
 * - known bugs
 * 
 */

# include "mint/mint.h"

# include "mint/asm.h"
# include "mint/dcntl.h"
# include "mint/file.h"
# include "mint/ioctl.h"
# include "mint/proc.h"
# include "mint/signal.h"
# include "mint/ssystem.h"
# include "mint/stat.h"
# include "cookie.h"

# include "libkern/libkern.h"

# include "loop.h"


/*
 * version
 */

# define VER_MAJOR	0
# define VER_MINOR	1
# define VER_STATUS	


/*
 * debugging stuff
 */

# if 1
# define DEV_DEBUG	1
# endif

# if 0
# define INT_DEBUG	1
# endif


/*
 * default settings
 */



/*
 * messages
 */

# define MSG_VERSION	str (VER_MAJOR) "." str (VER_MINOR) str (VER_STATUS) 
# define MSG_BUILDDATE	__DATE__

# define MSG_BOOT	\
	"\033p Loopback device driver version " MSG_VERSION " \033q\r\n"

# define MSG_GREET	\
	" " MSG_BUILDDATE " by Patrice Mandin.\r\n\r\n"

# define MSG_FAILURE	\
	"\7\r\nSorry, driver NOT installed - initialization failed!\r\n\r\n"


/****************************************************************************/
/* BEGIN kernel interface */

struct kerinfo *kernel;

/* END kernel interface */
/****************************************************************************/

/****************************************************************************/
/* BEGIN definition part */

#define MAX_LOOP_DEVICES 8

typedef struct {
	char dev_name[32];		/* Full device name, u:\\dev\loop<n> */
	char *bound_filename;	/* Name of file bound to loop device */
	int fd;					/* Its file descriptor */
} loopinfo_t;

/* Install time
 */
static ushort install_date;
static ushort install_time;

/*
 * device driver routines - top half
 */
static long _cdecl	loop_open	(FILEPTR *f);
static long _cdecl	loop_write	(FILEPTR *f, const char *buf, long bytes);
static long _cdecl	loop_read	(FILEPTR *f, char *buf, long bytes);
static long _cdecl	loop_lseek	(FILEPTR *f, long where, int whence);
static long _cdecl	loop_ioctl	(FILEPTR *f, int mode, void *buf);
static long _cdecl	loop_datime	(FILEPTR *f, ushort *timeptr, int rwflag);
static long _cdecl	loop_close	(FILEPTR *f, int pid);
static long _cdecl	loop_select	(FILEPTR *f, long proc, int mode);
static void _cdecl	loop_unselect	(FILEPTR *f, long proc, int mode);

/*
 * device driver map
 */

static DEVDRV raw_devtab =
{
	loop_open,
	loop_write, loop_read, loop_lseek, loop_ioctl, loop_datime,
	loop_close,
	loop_select, loop_unselect,
	NULL, NULL
};


/*
 * debugging stuff
 */

# ifdef DEV_DEBUG
#  define DEBUG(x)	KERNEL_DEBUG x
#  define TRACE(x)	KERNEL_TRACE x
#  define ALERT(x)	KERNEL_ALERT x
# else
#  define DEBUG(x)
#  define TRACE(x)
#  define ALERT(x)	KERNEL_ALERT x
# endif

# ifdef INT_DEBUG
#  define DEBUG_I(x)	KERNEL_DEBUG x
#  define TRACE_I(x)	KERNEL_TRACE x
#  define ALERT_I(x)	KERNEL_ALERT x
# else
#  define DEBUG_I(x)
#  define TRACE_I(x)
#  define ALERT_I(x)	KERNEL_ALERT x
# endif

/*
 * Gemdos kernel functions
 */

#define K_FOPEN		(*kernel->dos_tab[0x3d])
#define K_FCLOSE	(*kernel->dos_tab[0x3e])
#define K_FSEEK		(*kernel->dos_tab[0x42])
#define K_FREAD		(*kernel->dos_tab[0x3f])
#define K_FWRITE	(*kernel->dos_tab[0x40])

/* END definition part */
/****************************************************************************/

/****************************************************************************/
/* BEGIN global data definition & access implementation */

/*
 * global data structures
 */

static loopinfo_t loop_infos[MAX_LOOP_DEVICES];

static struct dev_descr raw_dev_descriptor =
{
	&raw_devtab,
	0,		/* dinfo -> fc.aux */
	0,		/* flags */
	NULL,	/* struct tty * */
	0,		/* drvsize */
	S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH,
	NULL,	/* bdevmap */
	0,		/* bdev */
	0		/* reserved */
};
	
char device_name[]="u:\\dev\\loop";

/* END global data & access implementation */
/****************************************************************************/

/****************************************************************************/
/* BEGIN initialization - top half */

DEVDRV * _cdecl
init (struct kerinfo *k)
{
	int len, i;
	char *name;

	kernel = k;
	
	c_conws (MSG_BOOT);
	c_conws (MSG_GREET);
	
	DEBUG (("%s: enter init", __FILE__));
	
	/* install it */
	len = strlen(device_name);
	for (i=0; i<MAX_LOOP_DEVICES; i++) {
		name = loop_infos[i].dev_name;
		strcpy(name, device_name);
		if (i<10) {
			name[len+1] = i % 10;
			name[len+2] = ' ';
		} else {
			name[len+1] = i / 10;
			name[len+2] = i % 10;
		}
		name[len+3] = '\0';
		raw_dev_descriptor.dinfo = i;

		loop_infos[i].bound_filename=NULL;
		loop_infos[i].fd=-1;

		DEBUG (("[LOOP] init: device [%s]", name));

		if (d_cntl (DEV_INSTALL, device_name, (long) &raw_dev_descriptor)<=0) {
			ALERT (("[LOOP] init: Unable to install %s", name));
		}		
	}

	install_time = timestamp;
	install_date = datestamp;

	return (DEVDRV *) 1L;
}

/* END initialization - top half */
/****************************************************************************/

/****************************************************************************/
/* BEGIN device driver routines - top half */

static long _cdecl
loop_open (FILEPTR *f)
{
	DEBUG (("loop_open [%i]: enter (%lx)", f->fc.aux, f->flags));

	f->pos = 0;

	DEBUG (("loop_open: return E_OK (added %lx)", f));
	return E_OK;
}

static long _cdecl
loop_close (FILEPTR *f, int pid)
{
	DEBUG (("loop_close [%i]: enter", f->fc.aux));
	
	return E_OK;
}


/* raw write/read routines
 */
static long _cdecl
loop_write (FILEPTR *f, const char *buf, long bytes)
{
	loopinfo_t *loopinfo;
	long done = 0;
	
	DEBUG (("loop_write [%i]: enter (%lx, %ld)", f->fc.aux, buf, bytes));
	
	loopinfo = &loop_infos[f->fc.aux];
	if (loopinfo->fd < 0) {
		DEBUG (("loop_write: no bound file"));
		return ENOSYS;
	}
	
	done = K_FWRITE(loopinfo->fd, bytes, buf);

	DEBUG (("loop_write: leave (%ld)", done));
	return done;
}

static long _cdecl
loop_read (FILEPTR *f, char *buf, long bytes)
{
	loopinfo_t *loopinfo;
	long done = 0;
	
	DEBUG (("loop_read [%i]: enter (%lx, %ld)", f->fc.aux, buf, bytes));
	
	loopinfo = &loop_infos[f->fc.aux];
	if (loopinfo->fd < 0) {
		DEBUG (("loop_read: no bound file"));
		return ENOSYS;
	}

	done = K_FREAD(loopinfo->fd, bytes, buf);
	
	DEBUG (("loop_read: leave (%ld)", done));
	return done;
}

static long _cdecl
loop_lseek (FILEPTR *f, long where, int whence)
{
	loopinfo_t *loopinfo;
	DEBUG (("loop_lseek [%i]: enter (%ld, %d)", f->fc.aux, where, whence));
	
	loopinfo = &loop_infos[f->fc.aux];
	if (loopinfo->fd < 0) {
		DEBUG (("loop_lseek: no bound file"));
		return ENOSYS;
	}

	return K_FSEEK(where, loopinfo->fd, whence);
}

static long _cdecl
loop_ioctl (FILEPTR *f, int mode, void *buf)
{
	loopinfo_t *loopinfo;
	long r = E_OK;
	
	DEBUG (("loop_ioctl [%i]: (%x, (%c %i), %lx)", f->fc.aux, mode, (char) (mode >> 8), (mode & 0xff), buf));

	loopinfo = &loop_infos[f->fc.aux];
	
	switch (mode)
	{
		case FIONREAD:
		case FIONWRITE:
		case FIOEXCEPT:
		{
			*(long *) buf = 0;
			return E_OK;
		}

		case LOOP_BIND:
		{
			long handle;
			int len;

			r = ENOSYS;

			/* File already bound to this device ? */
			if (loopinfo->fd >= 0) {
				break;
			}
				
			/* Bind file to this device */
			handle = K_FOPEN((char *)buf, 2); /* open read/write */
			if (handle<0) {
				r = handle;
				break;
			}

			/* Copy filename to local memory */
			len = strlen(buf)+1;
			loopinfo->bound_filename = kmalloc(len);
			if (loopinfo->bound_filename == NULL) {
				break;
			}
			strcpy(loopinfo->bound_filename, buf);
			loopinfo->bound_filename[len+1]='\0';

			loopinfo->fd = handle & 0xffff;
			r = E_OK;
			break;
		}
		case LOOP_UNBIND:
		{
			r = E_OK;
			/* No file bound to this device ? */
			if (loopinfo->fd < 0) {
				break;
			}

			/* Unbind file from this device */
			K_FCLOSE(loopinfo->fd);
			if (loopinfo->bound_filename) {
				kfree(loopinfo->bound_filename);
				loopinfo->bound_filename=NULL;
			}
			loopinfo->fd = -1;
			break;
		}
		
		default:
		{
			/* Fcntl will automatically call tty_ioctl to handle
			 * terminal calls that we didn't deal with
			 */
			r = ENOSYS;
			break;
		}
	}
	
	DEBUG (("loop_ioctl: return %li", r));
	return r;
}

static long _cdecl
loop_datime (FILEPTR *f, ushort *timeptr, int rwflag)
{
	DEBUG (("loop_datime [%i]: enter (%i)", f->fc.aux, rwflag));
	
	if (rwflag)
		return EACCES;
	
	*timeptr++ = install_time;
	*timeptr = install_date;
	
	return E_OK;
}

static long _cdecl
loop_select (FILEPTR *f, long proc, int mode)
{
	DEBUG (("loop_select [%i]: enter (%li, %i)", f->fc.aux, proc, mode));
	
	/* we're always ready for I/O */
	return 1;
}

static void _cdecl
loop_unselect (FILEPTR *f, long proc, int mode)
{
	DEBUG (("loop_unselect [%i]: enter (%li, %i)", f->fc.aux, proc, mode));

	/* nothing for us to do here */
}

/* END device driver routines - top half */
/****************************************************************************/
